# DataFrame 选择行用于反应式输入时会出错，需要从GitHub安装最新的shiny包,尚未发布。
# https://github.com/posit-dev/py-shiny/issues/676
# https://github.com/posit-dev/py-shiny/issues/739
# pip install https://github.com/posit-dev/py-shiny/tarball/main
# 导入包
# 处理数据、作图用的包
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from pathlib import Path
import time, sys, os
# 输出PDF房价分析报告用的包
import pickle
import nbformat
from nbconvert import PDFExporter
from nbconvert.preprocessors import ExecutePreprocessor, TagRemovePreprocessor
import concurrent.futures
import tempfile
# 这是Shiny for Python要用到的函数
from shiny import App, reactive, render, ui

# 运行另外一个目录下的Python脚本
sys.path.append('/home/jean/scripts')
t1 = time.time()
#加载所有的模型，以最佳参数一次过训练好所有模型。
import Melbourne_Regress as Melbourne
# 读取所有模型的结果为全局变量
train = Melbourne.PredictTrain()   # 训练集预测数据
valid = Melbourne.PredictValid()   # 测试集预测数据
perf  = Melbourne.performance()    # 各种算法性能数据
perf = perf.round(3)
perf.columns = ["算法","训练集","验证集","耗时"]
# 测试集的特征矩阵，用于显示异常房价的原始数据
valid_X = Melbourne.valid_X2
t2 = time.time()
# 加载模型总耗时
total = np.round(t2-t1,2)

# 算法模型列表
algoValues = train.columns
algoKeys = ["Origin","SVM","RandomForest","GBR","XGB","LigthGBM","CatBoost","Blend"]
algos = dict(zip(algoKeys, algoValues))

# 创建一个线程池执行器，用于后面另外启动一个线程生成PDF报告并下载
executor = concurrent.futures.ThreadPoolExecutor(max_workers=2)
future = None


# 定义UI函数，APP有两个输入参数：算法名称和异常值阈值%
app_ui = ui.page_fluid(
    # Javascript 处理父窗口传入的参数等。
    ui.tags.head(
        ui.tags.script(ui.HTML("""
                // 记录父窗口，初始为空。
                parent = null;
                // 处理接收到的消息。
                window.addEventListener('message', function(e) {
                        //alert(e.data);
                        try{
                            //记录父窗口以备回发信息
                            parent = e.source;
                            //向服务器发送input变量更新消息
                            Shiny.setInputValue('algo', e.data);
                        } catch (error){
                              alert(error);
                            }
                        },false);

                // 点击时传出参数    
                $(document).on('click', '.btn-success', function (evt) {
                      evt.preventDefault(); 
                      var selected = document.getElementById('rows_selected');
                      if (selected ==null || selected.value.length==0){
                         alert('没有选中的行!');
                      }else{
                        if (parent == null){
                          alert(selected.value);
                        } else {
                          try{
                            parent.postMessage(selected.value, '*');
                          } catch(error){
                            alert(error);
                          }
                        }
                      }
                });  
                
              """))
    ),
  
    ui.layout_sidebar(
        ui.panel_sidebar(
              ui.tags.h3("墨尔本房价回归模型示例"),
              ui.input_select(
                  'algo',
                  '回归算法',
                  algoKeys[1:],
                  selected = 'CatBoost'
              ),
              ui.input_slider(
                  "threshold",
                  "异常值阀值%：",
                  min = 20, 
                  max = 60, 
                  value = 30
              ),
              ui.HTML("<br>总耗时<br>"),
              ui.output_text("seconds"),
              ui.HTML("<br>模型性能<br>"),
              ui.output_table("performance"),
              ui.HTML("""
              说明：这是对ln(房价)，即房价自然对数的回归，上表的ACC指标也是如此，因为ln(房价)接近正态分布，这样回归精度更高。<br>
              exp(Predict)换算回原量纲后，回归精度略有下降，也接近88%。
              """),
              ui.HTML("<br><br>"),
              ui.input_text_area("rows_selected","已选择异常值行号", rows =6, resize="vertical", value =""),
              # 插入javascript，禁止自己修改 rows_selected input_text_area
              ui.tags.script(ui.HTML("""
                  var rows_selected = document.getElementById('rows_selected');
                  rows_selected.disabled = true;
              """)),
              ui.HTML("<br>"),
              ui.input_action_button("sendout", "传出选中的行", class_ = "btn-success"),
              ui.HTML("<br>"),
              ui.download_button("download", "下载分析报告"),
        ),
        ui.panel_main(
              ui.tags.h3(ui.output_text("parameters")),
              ui.HTML("<div style='text-align: center;'>"),
              ui.output_plot("train_plot", width='600px', height='400px'),
              ui.HTML("</div>"), 
              ui.HTML("<div style='text-align: center;'>"),
              ui.output_plot("valid_plot", width='600px', height='400px'),
              ui.HTML("</div>"),
              ui.HTML("<br><h3>测试集预测值异常（阀值以外区域）样本</h3><br>"),
              ui.output_data_frame("outliers"),
        ),
    ),
)

# matplotlib 作图中文设置
plt.rcParams["font.sans-serif"]=["SimHei"] #设置字体
plt.rcParams["axes.unicode_minus"]=False #正常显示负号

# 画训练集或测试集拟合效果图
def PlotPrediction(true,predicted, threshold=0.2, title = "数据集: "):
    df = pd.DataFrame({"Real":true,"Predicted":predicted})
    df.sort_values(by=["Real"], inplace=True)
    df = df.reset_index(drop=True)
    fig, ax = plt.subplots()
    plt.tick_params(labelsize=15)
    ax.set_title(title + '真实值 vs 预测值', fontsize=25)
    ax.plot(list(range(0,len(df["Predicted"]))),df["Predicted"],'b.',markersize=1,label='预测值')
    ax.plot(list(range(0,len(df["Real"]))),df["Real"],'r-',label='真实值') 
    ax.fill_between(list(range(0,len(df["Real"]))),df["Real"]+np.log(1-threshold),
                     df["Real"]+np.log(1+threshold),facecolor='yellow',interpolate=True)
    plt.xlabel("样本", fontsize=20)
    plt.ylabel("对数房价", fontsize=20)
    plt.legend(loc='upper left',fontsize=20);
    return fig    


# 定义服务器函数
def server(input, output, session):
    # 输出选择的算法与阈值
    @output
    @render.text
    def parameters():
        return f'算法：{input.algo()}      异常值阀值：{input.threshold()}%'

    # 输出总耗时
    @output
    @render.text
    def seconds():
        return f'{total} 秒'

    # 输出模型性能
    @output
    @render.table
    def performance():
        return perf

    # 画训练集拟合效果图
    @output
    @render.plot
    def train_plot():
        g = PlotPrediction(train["origin"], train[algos[input.algo()]], \
            float(input.threshold())/100, "训练集: ")
        return g

    # 画测试集拟合效果图
    @output
    @render.plot
    def valid_plot():
        g = PlotPrediction(valid["origin"], valid[algos[input.algo()]], \
            float(input.threshold())/100, "测试集: ")
        return g

    # 输出异常值列表
    @output
    @render.data_frame
    def outliers():
        # 根据算法和阈值标注异常值
        data = pd.DataFrame({"origin":valid["origin"], "predict":valid[algos[input.algo()]]})
        data["outlier"] = False
        data.loc[data["predict"] < data["origin"]+np.log(1-input.threshold()/100),["outlier"]] = True
        data.loc[data["predict"] > data["origin"]+np.log(1+input.threshold()/100),["outlier"]] = True
        # 选出要显示的列
        valid_X3 = valid_X[["Lattitude","Longtitude","Distance","BuildingArea","Landsize","YearBuilt","Year"]]
        # 处理房屋类型
        valid_X3["Type"] = valid_X["Type_h"]*4+valid_X["Type_t"]*2+valid_X["Type_u"]
        valid_X3.loc[valid_X3["Type"]==4,["Type"]] = "H"
        valid_X3.loc[valid_X3["Type"]==2,["Type"]] = "T"
        valid_X3.loc[valid_X3["Type"]==1,["Type"]] = "U"
        # 转换为原量纲
        valid_X3["Origin"] = np.round(np.exp(data["origin"]),2)
        valid_X3["Predict"] = np.round(np.exp(data["predict"]),2)
        # 选出测试集中的异常值行
        outliers = valid_X3[data["outlier"]==True]
        # 计算预测值的偏差并排序
        outliers["SE"] = np.round((outliers["Predict"] - outliers["Origin"])/outliers["Origin"],3)
        outliers.sort_values("SE", inplace = True)
        # 行号索引转换为列，以便后面对选出的异常值行跟进处理
        # 因原始数据读入为Pandas Data Frame后才划分训练集测试集，行号索引就是原始数据中的行号。 
        outliers["index"] = outliers.index
        # 放入 session 中以便后面选择行时引用。
        # https://github.com/posit-dev/py-shiny/issues/697
        # session中只有user、groups两个属性可以放值。用user，设置为字典可以放多个变量。
        # 可以用下面的程序打印session的结构：
        # for property, value in vars(session).items():
        #     print(property, ":", value)
        # 后面改了列名，所以要用copy()保存一份原列名的进session。session.user的初始值是None。
        session.user = {"outliers":outliers.copy()}
        outliers.columns = ["纬度","经度","距离CBD","建筑面积","占地面积","建成年份","出售年份",\
                  "类型", "真实值", "预测值", "偏差%", "行号"]
        # 渲染成Excel电子表格的格式
        return render.DataGrid(outliers, row_selection_mode="multiple", \
                  summary="正在查看{start} 到 {end} 行，一共 {total}行。")

    # 处理选中的行，选中的行命名的规则是 input.输出ID_selected_rows
    # 前面输出ui.output_data_frame("outliers")定义了ID为"outliers"，
    # 所以是input.outliers_selected_rows
    @reactive.Effect
    @reactive.event(input.outliers_selected_rows)
    def _():
        rows = input.outliers_selected_rows()
        # 从session中取出前面保存的当前输入下的异常值列表
        outliers = session.user["outliers"]
        if len(rows)>0:
            selected = outliers.iloc[list(rows),]
            rows2 = list(selected["index"])
            res = f'{rows2}'
            # 去掉列表的[]
            res = res[1:len(res)-1]
            # 去掉list格式化时所有的空格
            res = res.replace(" ","")
        else:
            res = ""
        # 更新到浏览器端的输入中显示
        ui.update_text_area(id = "rows_selected", value = res)
    

    # 更新由宿主页面传入的algo参数
    @reactive.Effect
    @reactive.event(input.algo)
    def _():
        print(input.algo())
        ui.update_select("algo",selected = input.algo())


    # 重新执行Jupyter Notebook，并转换为PDF输出
    def execute_and_convert_notebook_to_pdf(input_file_path, output_file_path):
        # 读入 notebook
        with open(input_file_path) as f:
            nb = nbformat.read(f, as_version=4)

        # 重新执行 notebook
        ExecutePreprocessor(timeout=None).preprocess(nb, {})
        
        # 添加过滤步骤，有些代码块不输出代码，或整个单元格不输出，或不显示输出。
        # https://nbconvert.readthedocs.io/en/latest/removing_cells.html
        tagremove_preprocessor = TagRemovePreprocessor()
        tagremove_preprocessor.remove_cell_tags = ("hide",)
        tagremove_preprocessor.remove_all_outputs_tags = ("hide-output",)
        tagremove_preprocessor.remove_input_tags = ("hide-input",)
        tagremove_preprocessor.remove_single_output_tags = ("hide-single",)
        tagremove_preprocessor.preprocess(nb, {})

        # Create PDF exporter
        nb.metadata = {'title' : '墨尔本房价分析报告', 'authors' : [{"name": "Jean Ye"}]}
        pdf_exporter = PDFExporter() # 默认模板的报告标题是"Notebook"
        ## https://stackoverflow.com/questions/68015894/nbconvert-6-how-to-customize-latex-template
        # pdf_exporter = PDFExporter(template_name='mytemplate')  # 引用自定义模板
        # pdf_exporter.exclude_input = True  # exclude input cells in the output PDF
        print(pdf_exporter.template_name) # 打印当前使用模板名称
        # Process and save to PDF
        pdf_data, _ = pdf_exporter.from_notebook_node(nb)
        with open(output_file_path, 'wb') as f:
            f.write(pdf_data)        


    # 下载pdf分析报告，函数名与ui.download_button()的id "download"相同。
    @session.download()
    def download():
        # 训练集拟合数据
        trainSelected = pd.DataFrame({"origin":train["origin"], "predict":train[algos[input.algo()]]})
        trainSelected.reset_index(inplace=True)
        trainSelected["index"] =  trainSelected.index
        # 测试集拟合数据
        validSelected = pd.DataFrame({"origin":valid["origin"], "predict":valid[algos[input.algo()]]})
        validSelected.reset_index(inplace=True)
        validSelected["index"] =  validSelected.index
        # 从session中取出前面保存的当前输入下的异常值列表
        outliers = session.user["outliers"]
        # 聚合成一个字典，以便pickle写入磁盘传给Jupyter Notebook
        parameters = {"algo":input.algo(),
                      "threshold":input.threshold(),
                      "perf":perf,
                      "trainSelected":trainSelected, 
                      "validSelected":validSelected,
                      "outliers":outliers,
                      "rows_selected":input.rows_selected()
                      }
        # Notebook报告的参数dump到磁盘文件中
        pickleFile = tempfile.NamedTemporaryFile(suffix=".pkl").name
        print(f'临时文件的名字：{pickleFile}')
        with open(pickleFile, "wb") as f:
            pickle.dump(parameters, f)
        print("Parameter dictionary dumped to: "+ str(pickleFile))
        # pickle文件名参数通过环境变量传给Jupyter Notebook，它从中加载参数。
        os.environ['pickleFile'] = str(pickleFile)

        # 这里调用nbconvert 生成PDF分析报告
        # construct filename
        inputFile = str(Path(__file__).parent / "report.ipynb")
        outputFile = tempfile.NamedTemporaryFile(suffix=".pdf").name
        print(f'临时文件的名字：{outputFile}')
        
        # 重新执行notebook并生成PDf分析报告
        id = ui.notification_show("正在生成分析报告......", duration=None)
        try:
            # 启动一个线程生成PDF报告
            global future
            future = executor.submit(execute_and_convert_notebook_to_pdf, inputFile, outputFile)
            # 等待PDF生成任务完成
            while future.running():
                time.sleep(1)    # 等待1秒
        except Exception as e:
            print(e)
        ui.notification_remove(id)
        # 删除pickle数据文件
        os.remove(pickleFile)
        # 应该增加下载后删除报告文件的选项，正如R Shiny那样。
        # https://github.com/posit-dev/py-shiny/issues/450
        return outputFile


# 组装成一个Shiny APP
app = App(app_ui, server)
